#
#-------------------------------------------------------------------------------
#  @(#) siemens-bss/parsersrc/SIEMENS_BSS.pm siemens-bss_r3.3.0.5:cset.003251:32:32@(#) 
#-------------------------------------------------------------------------------
#
#   Entry Points:   New, load_config, process
#
#   This module takes a raw file, and converts it to a PIF file. 
#   
#  Copyright (C) 2007 Vallent Software Systems UK. All rights reserved.
#
package SIEMENS_BSS;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

use AudLog;
use PIF_Handler;
use File::Basename;
use GenUtils;

use Data::Dumper;

require Exporter;
require AutoLoader;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw();
$VERSION = '0.01';


my @MONTHS = ( qw (Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec) );

my %seen = ();

################################################################################
# Subroutine name:  New()
#
# Description:      Object initialisation routine
#
# Arguments:        None
#
# Returns:          Object reference
#
sub New {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};

    # These variables should be set to true of false
    # depending on whether or not the desired type of output file
    # is wanted
    $self->{PRODUCE_PIF} = "True";
    $self->{DEFAULT_NULL_VALUE} = "NULL";

    bless ($self, $class);
}

################################################################################
# Subroutine name:  load_config()
#
# Description:      Object configuration loading routine
#
# Arguments:        keep_input_dir(scalar) - indicating where to store input file 
#                                         once it has been processed (if defined). 
#                   debug (scalar) - a boolean indicating whether or not the
#                                    parser is being run in debug mode.
#                   config (scalar) - a reference to a hash that contains all
#                                     the configuration options that have to
#                                     be loaded.
#
# Returns:          0 for success,
#                   the number of errors found for failure
#
sub load_config {
    my $self = shift;
    ($self->{keep_input_dir}, $self->{debug}, $self->{'__config__'}) = @_;
    
    # Inserting all the configuration information into the objects records
    my ($key, $num_errors,$ne);
    foreach $key ( keys %{$self->{'__config__'}} ) {
        $self->{$key} = $self->{'__config__'}->{$key};
    }
    
    $num_errors = 0;
    # CHECK MANDATORY OPTIONS are present
    my %mandatory = 
        (
         HEADER_FIELDS_FOR_PIF_FILENAME => 'ARRAY',
         MONITORED_OBJECTS              => 'HASH',
         OUTPUT_MEASURE_IDENTIFIER      => 'SCALAR',
         OBJECT_CLASSES                 => 'HASH',
         KEEP_VALIDITY_VALUES           => 'ARRAY',
         );
    my %optional = 
        (
         DEFAULT_NULL_VALUE             => 'SCALAR',
         ENABLE_PIF_CACHE               => 'SCALAR',
         ENABLE_FILENAME_TIMESTAMP      => 'SCALAR',
         );

    # mandatory and optional items for command config
    my %oc_mandatory = 
        (
         CLASS                          => 'SCALAR',
         MEASUREMENT_LIST               => [ 'ARRAY', 'SCALAR' ],
         MONITORED_OBJECTS_KEYS         => 'ARRAY',
         );
    my %oc_optional = 
        (
         OPTIONAL_VALUE_COUNTER         => 'SCALAR',
         OPTIONAL_COUNTER_AS_ROW_KEY    => 'SCALAR',
         OPTIONAL_COUNTER_AS_PIF_KEY   => 'SCALAR',
         INSERT_COUNTERS                => 'HASH',
         );
        
    # check the main configuration is okay
    $num_errors = &GenUtils::check_configuration($self,\%mandatory,\%optional);
    if ($num_errors) {
        LogMess("SIEMENS_BSS: $num_errors errors in configuration",3);
        return $num_errors;
    }

    # check the entries in command config.
    foreach (keys %{$self->{OBJECT_CLASSES}}){
        if ($ne = &GenUtils::check_configuration($self->{OBJECT_CLASSES}->{$_},
                                                 \%oc_mandatory,\%oc_optional)){
            LogMess("SIEMENS_BSS: $ne errors with configuration for object class $_",3);
        }
        $num_errors += $ne;
    }

    if ($num_errors){
        LogMess("SIEMENS_BSS: ERROR There are $num_errors errors with the configuration",2);
        return $num_errors;
    }
    $self->{statsData} = {};
    
    # build up a map from the class,counter name to the interested object names
    # so that mapping from configuration will be faster.
    $self->build_object_index();
    
    # map the list of allowed validity values into a hash
    # for quicker access
    foreach (@{$self->{KEEP_VALIDITY_VALUES}}){
        $self->{KEEP_VALIDITY_MAP}->{$_} = 1;
    }

    # add filename_suffix
    push @{$self->{filename_suffix}}, 'PM';
    if ( exists($self->{ENABLE_FILENAME_TIMESTAMP}) && ($self->{ENABLE_FILENAME_TIMESTAMP} =~ /True/i) ) {
        # add timestamp to filename suffix
        push @{$self->{filename_suffix}}, time();
    }

    return $num_errors;
}


############################################################
# Subroutine name:  process_file()
#
# Description:Controls the processing of all the files that 
#        match the INPUT_FILE_DESCRIPTION field in the config 
#        information
#
# Arguments: self (scalar) - a reference to a hash that contains
#                all the configuration options for this process.                       
#            filenm (scalar) - the location and filename of the 
#                file to be processed. 
#				 headers(scalar) - a reference to a hash that 
#                contains filename and directory name extractions.					
#
# Returns:   (scalar) - 0 = success
#				 -1 = Failed, rename input file to '.bad' 
#				 -2 = Failed, do not rename input file to '.bad'
#
sub process_file {
    my ($self, $filenm, $headers) = @_;
    my ($line,$lc,@fields);
    my ($class,$key,$mtype,$cname);
    my ($moid,$motype,$i,$index);
    my ($mo_config,$mov,%counter_h,%common_h,$j,$row_cnt,$obj_rc,$key);
    my (%pif_data_h,$moid_key,$mo_subkey,@mo_subkeys,@validity_list);
    my ($v,$k,$p,$obj_class_config,$counter_ref);
    my ($date,$time,$filekey,$datekey,$version,%granularity,$prev_cname);
    my ($first_row);
    my $measSet;
	
    my ($valid_no, $valid_class, $turntobad) = -1;

    # Establish the directory that outputted files will be stored in.
    my $int_dir = $self->{OUTPUT_DIR};


    # Read in the raw file
    $self->{FILE} = open_file($filenm);
    if ( $self->{FILE} < 0) {
        LogMess("SIEMENS_BSS: ERROR could not open raw file $filenm",2);
        return -1;
    }
   
    # Extract filename from the file reference
    my $fname = basename($filenm);
    LogMess("SIEMENS_BSS: Processing PM file $fname",5);
    
    $lc = 0;
    
    # Loop file contents
  LINE: 
    while (!eof($self->{FILE})){
        # ignore blank lines
        $line = readline($self->{FILE});
        chomp($line);
        $lc++;

        next LINE if ($line =~ /^$/);

        $row_cnt = 0;
        %counter_h = ();
        %common_h = ();
        @mo_subkeys = @validity_list = ();

        $index = 0;
        # split based on main field separator tabs
        @fields = split('\t',$line);

        my $found = 0;
        ####################################################
        # PARSE FIELD 1
        # get the class and type (X,Y) from field 1
        # could be in array position 0 or 1 
        for ($index=0; $index < 2; $index++){
            if (($class,$mtype) = $fields[$index] =~ /\((\d+),(\d+)\)/){
                $found = 1;
                last;
            }   
        }
        
        unless ($found){
            LogMess("SIEMENS_BSS: ERROR could not determine class " .
                    "and type on line $lc",1);
            # Fatal error, exit and return -1.
            #return -1;
        }
    
        # create the key from the class and type
        $key = join("-#-",$class,$mtype);

        if (exists($self->{OBJECT_INDEX}->{$key})){
            $cname = $self->{OBJECT_INDEX}->{$key};
        }
        elsif (exists($self->{OBJECT_INDEX}->{$class})){
            $key = $class;
            $cname = $self->{OBJECT_INDEX}->{$class};
        }
        else {
            LogMess("Ignoring $class,$mtype, not required (line $lc)",5);
            next LINE;
        }

        LogMess("Line $lc: Class ($class,$mtype) maps to name $cname",5);
        #####################################################
        # if it's a new class flush out the PIF date to help
        # memory manangement
        # Files can be very large therefore this is required
        unless (exists($self->{ENABLE_PIF_CACHE}) &&
                $self->{ENABLE_PIF_CACHE}){
            if (defined $prev_cname && $cname !~ /^$prev_cname$/){
                $self->write_pif_cache(\%pif_data_h,$headers,\%granularity,$version);
            }
        }
        
        $prev_cname = $cname;

        $obj_class_config = $self->{OBJECT_CLASSES}->{$cname};
        
        $index++;
        #######################################################
        # PARSE FIELD 2 - the date
        #
        ($date,$time,$datekey) = $self->convert_date($fields[$index]);
        unless (defined $date){
            LogMess("SIEMENS_BSS: ERROR line $lc, date $fields[$index] cannot " .
                    " be converted",1);
			$turntobad = 1;
            next LINE;
        }

        # now at field 3, after this all field positions relative
        $index++;
        
        #######################################################
        # PARSE FIELD 3 - the managed object identifier
        # 
        # The monitored object can have one of 3 formats
        # SHORT - ss7l {6 0}
        # MEDIUM - {bss:6 ss7l:7}
        # LONG - ss7l {6 0} {bss:6 ss7l:0}
        $moid = $fields[$index];
        if (($motype,$mov) = $moid =~ /^(\w+) \{(\d+)/){ # handles the long or short form
            unless (exists($self->{MONITORED_OBJECTS}->{$motype})){
                LogMess("SIEMENS_BSS: ERROR line $lc, no mapping for monitored " .
                        "object $motype in MONITORED_OBJECTS configuration",1);
				$turntobad = 1;
                next LINE;
            }
        
            $mo_config = $self->{MONITORED_OBJECTS}->{$motype};
            # extract the key counters 
            $common_h{$mo_config->[0]} = $mov;
            $moid_key = $mov;
            # loop to get the rest
            for ($i=1; $i < scalar @$mo_config; $i++){
                ($common_h{$mo_config->[$i]}) = $fields[$i+$index] =~ /^(\d+)/;
                $moid_key .= "-#-" . $1;
            }
            # skip over the fields which contained the moid
            # $index += $i;

            $index += ($i-1);

	    my $tmp_index = $index;
	    my $long_moid = '';
	   
	    # Processing the long format monitored object identifier
	    if ($fields[$tmp_index] =~ /\d+\}\s*\{(\w+):(\d+)\}*$/) {
		while ( $fields[$tmp_index] !~ /\}$/) {
			unless (defined($fields[$tmp_index])) {
				LogMess("SIEMENS_BSS: ERROR line $lc, incomplete data",1);
				$turntobad = 1;
				next LINE;
			}
			$long_moid .= $fields[$tmp_index] . " ";
			$tmp_index++;
		}	
		
		$long_moid .= $fields[$tmp_index];
		chop($long_moid);
		$long_moid =~ /(\w+:.+)/;
		$long_moid = $1;
	
                while ($long_moid =~ /(\w+:\d+)/g) { 
                    my $item = $1;
                    
		    $item =~ /(\w+):(.+)/;

                    my $moid_type = $1;
                    my $moid_val = $2;

                    $common_h{$moid_type} = $moid_val;
		}
	   }
	   elsif ( $fields[$index] =~ /^.*(\d+)-(\d+)-(\d+)-(\d+)/) {
            	my $ncc = $1;
		my $mcc = $2;
		my $lac = $3;
		my $cell_id = $4;
		LogMess ("NCC: $ncc, MCC: $mcc, LAC: $lac, cell: $cell_id", 5);
		$common_h{"CEL_ID"} = $cell_id;
		$index ++;
	   }
        }
        elsif (($motype,$mov) = $moid =~ /^\{(\w+):(\d+)/){ # MEDIUM format
            my @ca = ($mov);
            $moid_key = $mov;
            $index++;
            while ($fields[$index] !~ /^\d+$/){
				unless (defined($fields[$index])) {
            		LogMess("SIEMENS_BSS: ERROR line $lc, incomplete data",1);
					$turntobad = 1;
					next LINE;
				}
                ($motype,$mov) = $fields[$index] =~ /^(\w+):(\d+)/;
                push @ca, $mov;
                $moid_key .= "-#-" . $mov;
                $index++;
            }
            unless (exists($self->{MONITORED_OBJECTS}->{$motype})){
                LogMess("SIEMENS_BSS: ERROR line $lc, no mapping for monitored " .
                        "object $motype in MONITORED_OBJECTS configuration",1);
				$turntobad = 1;
                next LINE;
            }

            $mo_config = $self->{MONITORED_OBJECTS}->{$motype};

            # now we have all the counter values create the counters
            for ($i = (scalar @$mo_config)-1; $i >= 0; $i--){
                $common_h{$mo_config->[$i]} = pop @ca;
            }
        }
        else {
            LogMess("SIEMENS_BSS: ERROR could not map mo types and values from $fields[$index]",1);
	    	$turntobad = 1;
            next LINE;
        }

        # skip over any old format stuff if present
        # we know when we hit the version when it's digits only
        while (defined($fields[$index]) && $fields[$index] !~ /^(\d+)$/){
            $index++;
        }
        $version = $fields[$index];

        # skip symbolic name, not required.
        $index += 2;

        ####################################################
        # PARSE FIELD 6 the granularity
        # granularity needs to be saved on a per measurement 
        # class basis
        $granularity{$cname} = $fields[$index++];

        ####################################################
        # PARSE FIELD 7 - the object result number
        # 
        # The count here determines how many key and validity
        # fields are to be processed subsequently
        # This line will map to $obj_rc lines in the PIF
        $obj_rc = $fields[$index++];
        if ($obj_rc <= 0){
            LogMess("SIEMENS_BSS: ERROR Line $lc error object count of $obj_rc is invalid - ignoring",1);
			$turntobad = 1;
            next LINE;
        }
        


        ####################################################
        # PARSE FIELD 8 - the optional value list 
        # This can consist of a single digit value representing
        # celltype or a set of keys, such 2 target bts's
        # e.g.  2       tgtbts {12     19} tgtbts {12  65}
        # It can also be any of the 3 allowed formats of the previous 
        # monitored object (SHORT, MEDIUM or LONG)
        # hence the complicated processing
        # The number of keys to be extracted is determined by
        # the previous value. 
        $i = 0;

        while ($i < $obj_rc){

            # single value type measurement
            if ($fields[$index] =~ /^(\d+)\s*$/){
                # if there is nothing configured for this field then 
                # it can be ignored.
                if (exists($obj_class_config->{OPTIONAL_VALUE_COUNTER})){
                    $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER}} = $1;
                }
                if (exists($obj_class_config->{OPTIONAL_COUNTER_AS_ROW_KEY}) &&
                    $obj_class_config->{OPTIONAL_COUNTER_AS_ROW_KEY}){
                    LogMess("Adding optional counter key $obj_class_config->{OPTIONAL_VALUE_COUNTER}",5);
                    $mo_subkeys[$i] = $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER}};
                }
                $i++;
                $index++;
            }
            elsif ($fields[$index] =~ /(\w+) \{(\d+)/){ # handle short and long format
                # list of objects
                $motype = $1;
                $mov = $2;
                LogMess("Another mo sub type $motype",5);
                unless (exists($self->{MONITORED_OBJECTS}->{$motype})){
                    LogMess("SIEMENS_BSS: ERROR line $lc, no mapping for monitored " .
                            "object $motype in MONITORED_OBJECTS configuration",1);
					$turntobad = 1;
                    next LINE;
                }
                $mo_config = $self->{MONITORED_OBJECTS}->{$motype};
                $counter_h{$row_cnt}->{"ADJ_" . $mo_config->[0]} = $mov;
                    $mo_subkey = $mov;
                for ($j=1; $j < scalar @$mo_config; $j++){
                    ($counter_h{$row_cnt}->{"ADJ_" . $mo_config->[$j]}) = $fields[$j+$index] =~ /^(\d+)/;
                    $mo_subkey .= "-#-" . $1;
                }
                # insert the line from the file the PIF row was derived from
                $index += $j-1;

                if ( $fields[$index] =~ /.*\d+-\d+-\d+-(\d+)/ ) {
                    my $val = $1;
                    my $ctr = "ADJ_CELL_ID";
                    $counter_h{$row_cnt}->{$ctr} = $val;
                }

                LogMess("$lc: Adding key row $row_cnt key $mo_subkey field $fields[$index]",5);
                $mo_subkeys[$i] = $mo_subkey;
                $row_cnt++;
                $i++;
                
                LogMess("Skipping forward from $fields[$index]... to next valid",5);
                while (defined($fields[$index]) &&
                       ($fields[$index] !~ /.*\s*\w+ \{\d+/) && 
                       ($fields[$index] !~ /^[\d\s]+$/)){
                    $index++;
                }
                # make it easier to process the next loop
                $fields[$index] =~ s/^\w+:\d+\} //g;
            }
            elsif ($fields[$index] =~ /^\{(\w+):(\d+)/){ # MEDIUM format
                my @ca;
                my @field_list = split(' ',$fields[$index++]);
                foreach (@field_list) {

                    # extract the last field containing the $motype
                    ($motype,$mov) = $_ =~ /(\w+):(.+)\}/;
                    push @ca, $mov;
                    $mo_subkey = $mov;

                    unless (exists($self->{MONITORED_OBJECTS}->{$motype})){
                        LogMess("SIEMENS_BSS: ERROR line $lc, no mapping for monitored " .
                                "object $motype in MONITORED_OBJECTS configuration",3);
                        next LINE;
                    }
                
                    $mo_config = $self->{MONITORED_OBJECTS}->{$motype};
                
                    # now we have all the counter values create the counters
                    for (my $j = (scalar @$mo_config)-1; $j >= 0; $j--){
                        $counter_h{$i}->{"ADJ_" . $mo_config->[$j]} = pop @ca;
                    }
                    LogMess("$lc: Adding key row $row_cnt key $mo_subkey field $fields[$index]",5);
                    $mo_subkeys[$i++] = $mo_subkey;
                }
                $row_cnt++;
            }
            else {
                LogMess("SIEMENS_BSS: line $lc ERROR could not map optional value item (" . 
                        $fields[$index]. ")",1);
				$turntobad = 1;
                next LINE;
            }
        }

        # Enough data has been read at this stage to be able
        # to initialize the PIF file key
        # initialize the pif data hash unless it has already been 
        # seen the structure of the PIF hash is
        #
        # CLASS
        #      -> DATETIME-#-{MONITORED_OBJECTS_KEYS}
        #                  -> {OPTIONAL_VALUE_COUNTER} (optional)
        #                                            -> DATA_KEY
        # The first 3 keys (class, datetime and the monitored object)
        # keys map to a single file.
        # the final key maps to an individual row in the file.

        $filekey = join("-#-",$datekey,@common_h{@{$obj_class_config->{MONITORED_OBJECTS_KEYS}}});

        if ($obj_class_config->{OPTIONAL_COUNTER_AS_PIF_KEY}){
            $filekey .= "-#-" .  $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER}};
        }

        $pif_data_h{$cname}->{$filekey} = () unless (exists($pif_data_h{$cname}->{$filekey}));
 
        ####################################################
        # PARSE FIELD 9 - the validity list
        # There is a validity flag for each object in the 
        # optional value list. This validity flag applies
        # to the set of counters for that object
        #
        # LogMess("Validity fields $fields[$index] and first counter list $fields[$index+1]",5);
        # now get the validity list
        $i = 0;
        foreach (split(' ',$fields[$index++])){
            # save the validity to see if subsequent counter values 
            # should be discarded
            $validity_list[$i++] = $_;
        }
	
        ####################################################
        # PARSE FIELD measset - measset
	# added an output for MeasSet specific for BR8 data
        #
        $measSet=$fields[$index++];
        #LogMess("measSet: $measSet",0);
	# allows setting this fiels as an iptional value in EngineConfig
        if (exists($obj_class_config->{OPTIONAL_VALUE_COUNTER2})){
            $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER2}} = $measSet;
        }
        if (exists($obj_class_config->{OPTIONAL_COUNTER2_AS_ROW_KEY}) &&
            $obj_class_config->{OPTIONAL_COUNTER2_AS_ROW_KEY}){
            LogMess("Adding optional counter key $obj_class_config->{OPTIONAL_VALUE_COUNTER2}",3);
	    LogMess(" setting key as: $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER2}}",5);
            $mo_subkeys[$row_cnt] = $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER2}};
	    $row_cnt++;
        }
        if ($obj_class_config->{OPTIONAL_COUNTER2_AS_PIF_KEY}){
            $filekey .= "-#-" .  $common_h{$obj_class_config->{OPTIONAL_VALUE_COUNTER2}};
        }
        $pif_data_h{$cname}->{$filekey} = () unless (exists($pif_data_h{$cname}->{$filekey}));


        ####################################################
        # PARSE FIELD 10 - the counter list 
        # The counters are stored as a list of value separated 
        # by blanks. Note this list of counters may translate
        # into a number of rows in the PIF.
        #

        # calculate how many counters apply to each object
	
        my @cv = split(' ',$fields[$index]);
	my $counter_subgroup = int(scalar @cv / $obj_rc );
	
        if ($counter_subgroup == 0) {
		if ( ($class != 1) && (($mtype != 60)|| ($mtype != 61)) ) {
			LogMess("SIEMENS_BSS: ERROR line $lc, FILE $fname, $cname $obj_rc invalid counter count " . 
			scalar @cv . " found in list " . $fields[$index],1);
			$turntobad = 1;
			next LINE;
		}
        }
        LogMess("Counter subgroup length is $counter_subgroup $cname obj count $obj_rc ",5);
        
        # map the counter values into the correct row and counter name in output
        # hash
        # the outer loop maps to each of the PIF rows and the
        # inner loop maps to the counters for that row.
        $j = 0;
        for ($i = 0; $i < $obj_rc; $i++){
            $p = 0;
            # if there is only one counter index begins at 0
            $k = ($counter_subgroup == 1) ? 0 : 1;
            my $counter_name = "C" . $class . "_" . $mtype . "_" ;
            while ($p < $counter_subgroup){
	    	# Bug #57057 - Special handling for class 5 data with measSet value 2
	    	# From C5_<mtype>_12 - C5_<mtype>_22 to C5_<mtype>_23 - C5_<mtype>_33
	    	# Bug #57243 - Added special handling for class 3 data with measSet value 2
	    	# From C3_<mtype>_12 - C3_<mtype>_22 to C3_<mtype>_23 - C3_<mtype>_33
		   	my $new_cname;
	    	if ((($class == 5) || ($class == 3)) && $measSet == 2 && $k >= 12){
			my $new_counter_ofs = 11;
			$new_cname =  $counter_name . ($k + $new_counter_ofs);
			$k++;
		} else {
	    		$new_cname =  $counter_name . $k++;
		}
	        $counter_h{$i}->{ $new_cname } = $cv[$j++];
        	$p++;
            }            
        }

        # Bug #57057 - Special handling for class 5 data with measSet value 2
        # Rename counters:
        # From C5_<mtype>_12 - C5_<mtype>_22 to C5_<mtype>_23 - C5_<mtype>_33
        #if ($class == 5 && $measSet == 2){
        #    my $cnt_name_prefix = "C" . $class . "_" . $mtype . "_" ;
        #    my $new_counter_ofs = 11;
        #    my $old_counter;
        #    my $new_counter;
        #    for ($k = 12; $k <= 22; $k++){
        #        for ($i = 0; $i < $obj_rc; $i++){
                     # Copy out the value to the new counter name
        #            $old_counter = $cnt_name_prefix . $k;
        #            $new_counter = $cnt_name_prefix . ($k + $new_counter_ofs);
        #            $counter_h{$i}->{$new_counter} =
        #                $counter_h{$i}->{$old_counter};
		    
                    # Delete the old counter hash
        #            delete $counter_h{$i}{$old_counter};
        #        }
        #    }
        #}


        ####################################################
        # 
        # Line is now read, time to validate and save the data
        #
        ####################################################
        # process each line to be written to the PIF
        # 1. Check the validity flag is okay, so the counter values
        # can be kept
        # 2. Create the full key and save the data.
      COUNTER_SET:
        foreach $i (keys %counter_h){
            my $moid_fullkey = join("-#-",$moid_key,$mo_subkeys[$i]);
            
            if (!exists($pif_data_h{$cname}->{$filekey}->{$moid_fullkey})){
                LogMess("index $i First time seeing key $moid_fullkey,$filekey for class $cname",5);
                %{$pif_data_h{$cname}->{$filekey}->{$moid_fullkey}} = ();
                $counter_ref = $pif_data_h{$cname}->{$filekey}->{$moid_fullkey};
                # first time seeing counters then add it to the list
                @$counter_ref{keys %common_h} = values %common_h;
                $first_row = 1;
            }
            else {
                $counter_ref = $pif_data_h{$cname}->{$filekey}->{$moid_fullkey};
                $first_row = 0;
            }                
            
            if (exists($self->{KEEP_VALIDITY_MAP}->{$validity_list[$i]})){
                # map the counter values
                @$counter_ref{keys %{$counter_h{$i}}} = values %{$counter_h{$i}};
            }
            elsif ($obj_class_config->{IGNORE_INVALID_VALUES}){
                LogMess("SIEMENS_BSS: ($moid_fullkey) Ignoring row for " . 
                        "$cname ($class,$mtype) due to invalid " .
                        "validity value $validity_list[$i] ($lc)",3); 
                if ($first_row){
                    LogMess("Deleting entry..",5);
                    $counter_ref = ();
                    delete($pif_data_h{$cname}->{$filekey}->{$moid_fullkey});
                }
                next COUNTER_SET;
            }
            else {
                LogMess("SIEMENS_BSS: ($moid_fullkey) Replacing counters values for " . 
                        "$cname ($class,$mtype) with null due to invalid " .
                        "validity value $validity_list[$i] ($lc)",3);
                foreach (keys %{$counter_h{$i}}){
                    $counter_ref->{$_} = 
                        ($_ =~ /^C\d+/) ? $self->{DEFAULT_NULL_VALUE} : $counter_h{$i}->{$_};
                }
                
            }
	
	    # Catch 3 different suspect flag situations
	    # 1st is the suspect flag for all counter type e.g SUSPECT_C_1
	    # 2nd is the last suspect flag for one measuremnt eg SUSPECT (for type 1)
	    # 3rd is the maximum suspect flag for one measuremnt eg SUSPECT_MAX_C1
	    if ($valid_no == -1 && $valid_class == -1){
		$valid_no = $validity_list[$i];
		$valid_class = $class;
	    }

            # Insert the SUSPECT flag into each PIF file
	    # Commented out the SUSPECT FLAG that is created for each row of data
	    # This is not currently needed but can be commented back in if required
            #$counter_ref->{"SUSPECT_C". $class . "_" . $mtype} = $validity_list[$i];
            $counter_ref->{SUSPECT} = $validity_list[$i];
	    $counter_ref->{"SUSPECT_MAX_C" . $class } = $valid_no;

	    if($valid_no <  $validity_list[$i] && $valid_class eq $class ){
	
	    	$valid_no = $validity_list[$i];
	     
	    }
	    elsif($valid_class ne $class){
		$valid_class = $class;
	        $valid_no = 0;
	    }
            # insert any required counters for this class
            if (exists($obj_class_config->{INSERT_COUNTERS})){
                while (($k,$v) = each %{$obj_class_config->{INSERT_COUNTERS}}){
                    $counter_ref->{$k} = $v;
                }
            }
            
            if ($self->{OUTPUT_MEASURE_IDENTIFIER}){
                $counter_ref->{"measIdentifier_C" . $class . "_" . $mtype} = '"' . $fields[0] . '"';
            }
            
            
        } # end loop to process each PIF row.

    } # end while to read lines
   
    close(FILE);

    LogMess("Completed processing file $filenm",5);

    # write out any remaining classes which had been parsed.
    $self->write_pif_cache(\%pif_data_h,$headers,\%granularity,$version);
    if( $turntobad == 1){
	return -1;
    }
    return 0;
}

############################################################
#
# Sub: build_object_index
#
# Description: builds up a simple reverse index from OBJECT_CLASSES
# configuration so that a class name, measurement type key may 
# be looked up quickly to see if it is configured, rather than
# having to check it against every entry in the configuration.
#
# Parameters: None
#
# Returns: Nothing.
############################################################
sub build_object_index {
    my ($self) = @_;
    my ($class,$config,$key);
    $self->{OBJECT_INDEX} = ();

    LogMess("SIEMENS_BSS: Building internal index for classes",5);

    while (($class,$config) = each %{$self->{OBJECT_CLASSES}}){

        if (ref($config->{MEASUREMENT_LIST}) eq "ARRAY"){
            foreach (@{$config->{MEASUREMENT_LIST}}){
                $key = join("-#-",$config->{CLASS},$_);
                $self->{OBJECT_INDEX}->{$key} = $class;
            }
        }
        elsif ($config->{MEASUREMENT_LIST} =~ /\*/){
            LogMess("Class $class applies to all measurements of class $config->{CLASS}",5);
            $self->{OBJECT_INDEX}->{$config->{CLASS}} = $class;
        }
        else {
            $key = join("-#-",$config->{CLASS},$config->{MEASUREMENT_LIST});
            $self->{OBJECT_INDEX}->{$key} = $class;
        }
    }

    my $oc =  scalar keys %{$self->{OBJECT_INDEX}};
    LogMess("Build up class/measurement index with $oc keys",5);
}


############################################################
#
# sub: write_pif_cache
#
# Desc: Writes out the PIF data cache, creating the necessary
# PIF files and writing out all counter values to each file.
# The cache is then cleared.
#
# Params: $1 - reference to the hash containing the data
#         $2 - reference to the headers from the FILENAME_HEADER_FIELDS.
#         $3 - the software version (from field 5)
#
# Returns: nothing
#
############################################################
sub write_pif_cache
{
    my ($self,$pif_data_ref,$headers,$granularity,$version) = @_;
    my ($cname,$h_ref,$h_ref2,$c_ref,$filekey);
    my (%header_h,$k,$v,@c_names,$pif_obj,$pif_ctx,$moid_key);
    my ($counter_ref,$config_ref,@fka,$i);
    my $all_counter_key;

    LogMess("Clearing out PIF data cache ",3);

    # get the pif object
    ($pif_obj,$pif_ctx) = GenUtils->get_pif_object();

    ########################################################
    #
    # create and write all the PIFs
    # Loop through each class type and then each output 
    # row which is written out to the PIF
    # 
    CLASS: while ( ($cname,$h_ref) =  each %$pif_data_ref){
        LogMess("Creating pif files for object class $cname",5);
        $config_ref = $self->{OBJECT_CLASSES}->{$cname};

        # process each date within that class
        CREATE_PIF: while (($filekey,$h_ref2) = each %$h_ref){
            
            $header_h{VERSION} = $version;
            $header_h{GRANULARITY} = $granularity->{$cname};
            @fka =  split("-#-",$filekey);
            $header_h{ENDDATE} = $fka[0];

            # add in the colon
            $fka[1] =~ s/(\d{2})(\d{2})/$1:$2/;
            $header_h{ENDTIME} = $fka[1];

            $i = 2;
            # the first 2 values contain the date and time
            # any other objects are header keys configured
            # in the monitored objects keys for the class
            foreach (@{$config_ref->{MONITORED_OBJECTS_KEYS}}){
                $header_h{$_} = $fka[$i++];
            }

            # add any filename counters
            while (($k,$v) = each %$headers){
                if ($v =~ /^(\d{4})(\d{2})(\d{2})$/){
                    $header_h{$k} = $3 .  $MONTHS[$2-1] . $1;
                }
                elsif ($k =~ /TIME/ && $v =~ /^(\d{2})(\d{2})$/){
                    $header_h{$k} = $1 . ":" . $2;
                }
                else {
                    $header_h{$k} = $v;
                }
            }

            unless (keys %{$h_ref2}){
                LogMess("No data rows - $filekey not creating PIF",4);
                next CREATE_PIF;
            }

            my $filename = '';

            unless ($filename) {
                # create the filename
                $filename = join("-#-",$cname,
                          @header_h{@{$self->{HEADER_FIELDS_FOR_PIF_FILENAME}}},
                          $filekey,@{$self->{filename_suffix}},"I.pif");
         
		if ( exists($self->{ENABLE_FILENAME_COUNTER}) && ($self->{ENABLE_FILENAME_COUNTER} =~ /True/i) ) {
			if ($seen{$filename}) {
				$filename =~ s/(-#-I.pif)$/-#-$seen{$filename}$1/g;
			}
			$seen{$filename}++;
		}
            }    
          
            $pif_obj->Open_PIF_Write($pif_ctx,$self->{OUTPUT_DIR} . '/' . $filename);
            
            $pif_obj->PIF_Write($pif_ctx,'HEADER',\%header_h);

	    # Bug #57242 - Create counter hash for this PIF
            my $all_counter_key = {};
            while (($moid_key,$counter_ref) = each %{$h_ref2}) {
                foreach (keys %{$counter_ref}) {
                    $all_counter_key->{$_} = '';
                }
            }

            # write out each row for this PIF
            while (($moid_key,$counter_ref) = each %{$h_ref2}){
                # delete any key counters
                foreach (@{$config_ref->{MONITORED_OBJECTS_KEYS}}){
                    delete $counter_ref->{$_};
                }
		
		# Bug #57242 - Assign the counter value as empty if the counter key not found 
		# in the hash
		foreach (keys %{$all_counter_key}) {
			if (!exists($counter_ref->{$_})) {
				$counter_ref->{$_} = '';
			}
		}

                $pif_obj->PIF_Write($pif_ctx,$cname,\%{$counter_ref});
                $c_ref = $counter_ref;
            }
            
            $pif_obj->Close_PIF_Write($pif_ctx);
            push @c_names, keys %$c_ref;
        }
    }

    %{$self->{stats_data}} = %header_h;
    &Counter_Statistics::add_raw_counters(\@c_names);

    undef %{$pif_data_ref};
}

############################################################
#
# sub: convert_date
#
# Description: converts the 2 siemens date format into the
# standard PIF/LIF format
#
# Parameters: $1 - the siemens date/time
#
# Returns: $1 - the date
#          $2 - the time
#          $3 - the date and time as a key
############################################################
sub convert_date
{
    my ($self,$datef) = @_;
    my ($date,$time,$dtk);
    # convert the date to proper format 
    if ($datef =~ /\w{3} (\w{3}) (\d{2}) (\d{2}):(\d{2}):\d{2} (\d{4})/){
        $date = $2 . $1 . $5;
        $time = $3 . ":" . $4;
        $dtk = $date . "-#-" . $3 . ":" . $4;
    }
    elsif ($datef =~ /(\d{2})\/(\d{2})\/(\d{4})\s+(\d{2}):(\d{2})/){
        $date = $1 . $MONTHS[$2-1] . $3;
        $time = $4 . ":" . $5;          
        $dtk = $date . "-#-" . $4 . ":" . $5;        
    }
    else {
        LogMess("SIEMENS_BSS: ERROR: DATETIME $datef cannot be parsed",5);
        undef $date;
    }
 
    return ($date,$time,$dtk);
}

################################################################################
#
# Subroutine name:  getHeaderStatsData
#
# Description:      Fills out the header statistics information so that it can
#                   be used by the stats processsing. If you have stats switched
#                   on for your module you must fill this function out.
#
# Arguments:        a reference to the statistics object to fill out
#
# Returns:          None
#
################################################################################
sub getHeaderStatsData {
    my ($self,$stats_ref) = @_;
    LogMess("getHeaderStatsData, filling statistics information ",5);
    
    my ($stats_name,$counter_name);
    
    # the header information is usually derived from the pif file
    # header, being stored in $self->{headerData}
    while (($stats_name,$counter_name) = each %{$stats_ref}){
        if (exists( $self->{stats_data}->{$counter_name})){
            $stats_ref->{$stats_name} = $self->{stats_data}->{$counter_name};
        }
        elsif ($counter_name eq "Type"){
            $stats_ref->{$stats_name} = "BSS";
        }
        else {
            LogMess("SIEMENS_BSS: Warning - could not find entry $counter_name" .
                    " in header data for statistics ",3);
            $stats_ref->{$stats_name} = "";
        }
    }
            
}

1;  # tells perl that the package is ready to run
